package RVVote::Controller::Evaluation;
use Mojo::Base 'Mojolicious::Controller';
use YAML;

sub process {
    my $c = shift;

    my @ballots = ();

    my $r;

    BALLOT:
    foreach my $ballot ( @{ $c->req->json } ) {

        my @choices = ();
        my %count   = ();
        
        CHOICE:
        foreach my $choice ( split /\W+/, $ballot->{value} ) {
            next CHOICE if $choice !~ /\w/; 
            if ( ! $count{ $choice }++ ) {
                push @choices, $choice;
            }
        }

        next BALLOT if ! scalar @choices; 

        push @ballots, {
            voter   => $ballot->{voter},
            choices => \@choices,
        }
    }

    my $total = ( scalar @ballots );
    my $half  = $total / 2;
    my @log;

    $c->stash->{ballots} = \@ballots;

    # primarni rozdeleni listku a pocitani celkoveho poctu 

    BALLOT:
    foreach my $ballot ( @ballots ) {
        my $choice = $ballot->{choices}[0];
        push @{ $r->{ $choice }{ballots} }, $ballot;

        CHOICE:
        foreach $choice ( @{ $ballot->{choices} } ) {
            $r->{ $choice }{votes}++;
            $r->{ $choice }{ballots} //= [];
        }
    }

    # oznacovani moznosti bez polovicni podpory

    CHOICE:
    foreach my $choice ( keys %{ $r } ) {
        if ( $r->{$choice}{votes} < $half  ) {
            $r->{$choice}{unsupported} = 1;
        }
    }

    push @log, {
        title => "Počáteční rozdělení ($total listků)", 
        data  => Dump $r, 
    }; 

    # vyrazeni moznosti kteri nemaji nadpolovicni podporu 
    CHOICE:
    foreach my $choice ( keys %{ $r } ) {
        if ( $r->{$choice}{unsupported}  ) {
            $r = transfer($r, $choice);
        }
    }

    push @log, {
        title => 'Po vyřazení možnosti bez nadpoloviční podpory', 
        data  => Dump $r, 
    }; 

    my @winners = winners( $r, $half );
    my $round = 1;

    ROUND:
    while ( ! scalar @winners && scalar keys %{ $r } > 1) {
        my $random = 0;

        my @loosers = loosers($r);

        if ( scalar @loosers == scalar keys %{$r}) {
            $random = 1;
            @loosers = ( $loosers[ int(rand(scalar @loosers))]);
        }

        LOOSER:
        foreach my $looser ( @loosers ) {
            $r = transfer($r, $looser); 
        }

        push @log, {
            title   => 'Vyřazovací kolo ' . $round++, 
            data    => Dump $r, 
            loosers => \@loosers,
            random  => $random,
        }; 

        @winners = winners( $r, $half );
    }

    $c->stash->{log} = [ map {
        { title => $_->{title}, data => Load( $_->{data}) }
    } @log ];

    $c->stash->{winners} = \@winners;        

    $c->render( template => 'result' );

}

sub transfer {
    my $r         = shift;
    my $choice    = shift;

    BALLOT: 
    foreach my $ballot ( @{ $r->{$choice}{ballots} } ) {
         HEIR:
         foreach my $heir ( @{ $ballot->{choices} } ) {
            next HEIR if $heir eq $choice;
            if ( exists $r->{ $heir } ) {
                push @{ $r->{ $heir }{ballots} }, $ballot;
                last HEIR;
            }
         }                                
    }
    delete $r->{ $choice };
    return $r ;

}

sub choices_by_ballots {
    my $r = shift;
    return sort { scalar @{$r->{$b}{ballots}} <=> scalar @{$r->{$a}{ballots}} } keys %{ $r }
}

sub winners {
    my $r    = shift;
    my $half = shift;

    my @candidates = ();
    my @winners    = ();

    CHOICE:
    foreach my $choice ( choices_by_ballots( $r ) ) {
        if ( scalar @{$r->{$choice}{ballots}} > $half ) {
            push @candidates, $choice;
        }        
    }

    my $max = 0;

    CANDIDATE:
    foreach my $choice ( sort { $r->{$b}{votes} <=> $r->{$a}{votes} } @candidates ) {
        $max ||= $r->{$choice}{votes}; 
        push @winners, $choice if $r->{$choice}{votes} == $max;
    }

    return @winners;
}

sub loosers {
    my $r    = shift;

    my @candidates = ();
    my @loosers    = ();

    my $min = 0;

    CHOICE:
    foreach my $choice ( reverse choices_by_ballots( $r ) ) {
        my $ballots = scalar @{$r->{$choice}{ballots}};
        $min ||= $ballots; 
        push @candidates, $choice if  scalar $ballots == $min;
    }

    $min = 0;

    CANDIDATE:
    foreach my $choice ( sort { $r->{$a}{votes} <=> $r->{$b}{votes} } @candidates ) {
        $min ||= $r->{$choice}{votes}; 
        push @loosers, $choice if $r->{$choice}{votes} == $min;
    }

    return @loosers;
}


1;
__END__

=cut


    ROUND:
    while ( 1 ) {

        push @{ $c->stash->{log} }, {
            title => $round++ . ". kolo" , 
            data  => { %{ $r } }, 
        }; 

        my ($min, $max, @loosers, $transfers); 

        CHOICE:
        foreach my $choice ( keys %{ $r } ) {
            my $count = scalar @{ $r->{$choice}{ballots} };                           

            if ( ! defined $max || $count > $max ) {
                $max = $count 
            }    

            if ( ! defined $min || $count < $min ) {    
                $min = $count;
                @loosers = ($choice); 
            }
            elsif ( defined $min && $count == $min ) {    
                push @loosers, $choice; 
            }
        }

        last ROUND if $max > $half;

        my $looser = (sort { $r->{$a}{votes} <=> $r->{$b}{votes} } @loosers)[0];  

        push @{ $c->stash->{log} }, {
            title => "Vyřazuje se možnost $looser", 
        }; 

        my $transfer;
        ($transfer, $r) = transfer($r, $looser); 
        last ROUND if ! $transfer;
    }

    my $winner = (choices_by_ballots( $r ))[0];
=cut
